@@ -1,5 +1,5 @@
 /*!
- * Copyright (c) 2018-2020 by Contributors
+ * Copyright (c) 2018-2021 by Contributors
  * \file code_folding_util.h
  * \author Hyunsu Cho
  * \brief Utilities for code folding
@@ -7,8 +7,8 @@
 #ifndef TREELITE_COMPILER_COMMON_CODE_FOLDING_UTIL_H_
 #define TREELITE_COMPILER_COMMON_CODE_FOLDING_UTIL_H_
 
-#include <dmlc/logging.h>
 #include <fmt/format.h>
+#include <treelite/logging.h>
 #include <unordered_map>
 #include <queue>
 #include <set>
@@ -24,7 +24,8 @@ namespace treelite {
 namespace compiler {
 namespace common_util {
 
-template <typename OutputFormatFunc>
+
+template <typename ThresholdType, typename LeafOutputType, typename OutputFormatFunc>
 inline void
 RenderCodeFolderArrays(const CodeFolderNode* node,
                        bool quantize,
@@ -36,12 +37,12 @@ RenderCodeFolderArrays(const CodeFolderNode* node,
                        std::string* array_cat_begin,
                        std::string* output_switch_statements,
                        Operator* common_comp_op) {
-  CHECK_EQ(node->children.size(), 1);
+  TREELITE_CHECK_EQ(node->children.size(), 1);
   const int tree_id = node->children[0]->tree_id;
   // list of descendants, with newly assigned ID's
   std::unordered_map<ASTNode*, int> descendants;
   // list of all OutputNode's among the descendants
-  std::vector<OutputNode*> output_nodes;
+  std::vector<OutputNode<LeafOutputType>*> output_nodes;
   // two arrays used to store categorical split info
   std::vector<uint64_t> cat_bitmap;
   std::vector<size_t> cat_begin{0};
@@ -57,16 +58,16 @@ RenderCodeFolderArrays(const CodeFolderNode* node,
     while (!Q.empty()) {
       ASTNode* e = Q.front(); Q.pop();
       // sanity check: all descendants must have same tree_id
-      CHECK_EQ(e->tree_id, tree_id);
+      TREELITE_CHECK_EQ(e->tree_id, tree_id);
       // sanity check: all descendants must be ConditionNode or OutputNode
       ConditionNode* t1 = dynamic_cast<ConditionNode*>(e);
-      OutputNode* t2 = dynamic_cast<OutputNode*>(e);
-      NumericalConditionNode* t3;
-      CHECK(t1 || t2);
+      OutputNode<LeafOutputType>* t2 = dynamic_cast<OutputNode<LeafOutputType>*>(e);
+      NumericalConditionNode<ThresholdType>* t3;
+      TREELITE_CHECK(t1 || t2);
       if (t2) {  // e is OutputNode
         descendants[e] = new_leaf_id--;
       } else {
-        if ( (t3 = dynamic_cast<NumericalConditionNode*>(t1)) ) {
+        if ( (t3 = dynamic_cast<NumericalConditionNode<ThresholdType>*>(t1)) ) {
           ops.insert(t3->op);
         }
         descendants[e] = new_node_id++;
@@ -76,7 +77,7 @@ RenderCodeFolderArrays(const CodeFolderNode* node,
       }
     }
     // sanity check: all numerical splits must have identical comparison operators
-    CHECK_LE(ops.size(), 1);
+    TREELITE_CHECK_LE(ops.size(), 1);
     *common_comp_op = ops.empty() ? Operator::kLT : *ops.begin();
   }
 
@@ -89,38 +90,33 @@ RenderCodeFolderArrays(const CodeFolderNode* node,
     std::string threshold;
     int left_child_id, right_child_id;
     unsigned int split_index;
-    OutputNode* t1;
-    NumericalConditionNode* t2;
+    OutputNode<LeafOutputType>* t1;
+    NumericalConditionNode<ThresholdType>* t2;
     CategoricalConditionNode* t3;
 
     std::queue<ASTNode*> Q;
     Q.push(node->children[0]);
     while (!Q.empty()) {
       ASTNode* e = Q.front(); Q.pop();
-      if ( (t1 = dynamic_cast<OutputNode*>(e)) ) {
+      if ( (t1 = dynamic_cast<OutputNode<LeafOutputType>*>(e)) ) {
         output_nodes.push_back(t1);
         // don't render OutputNode but save it for later
       } else {
-        CHECK_EQ(e->children.size(), 2U);
+        TREELITE_CHECK_EQ(e->children.size(), 2U);
         left_child_id = descendants[ e->children[0] ];
         right_child_id = descendants[ e->children[1] ];
-        if ( (t2 = dynamic_cast<NumericalConditionNode*>(e)) ) {
+        if ( (t2 = dynamic_cast<NumericalConditionNode<ThresholdType>*>(e)) ) {
           default_left = t2->default_left;
           split_index = t2->split_index;
           threshold
            = quantize ? std::to_string(t2->threshold.int_val)
                       : ToStringHighPrecision(t2->threshold.float_val);
         } else {
-          CHECK((t3 = dynamic_cast<CategoricalConditionNode*>(e)));
+          TREELITE_CHECK((t3 = dynamic_cast<CategoricalConditionNode*>(e)));
           default_left = t3->default_left;
           split_index = t3->split_index;
           threshold = "-1";  // dummy value
-          CHECK(!t3->convert_missing_to_zero)
-            << "Code folding not supported, because a categorical split "
-            << "is supposed to convert missing values into zeros, and this "
-            << "is not possible with current code folding implementation.";
-          std::vector<uint64_t> bitmap
-            = GetCategoricalBitmap(t3->left_categories);
+          std::vector<uint64_t> bitmap = GetCategoricalBitmap(t3->matching_categories);
           cat_bitmap.insert(cat_bitmap.end(), bitmap.begin(), bitmap.end());
           cat_begin.push_back(cat_bitmap.size());
         }
@@ -165,7 +161,7 @@ RenderCodeFolderArrays(const CodeFolderNode* node,
   }
   // 4. Render switch statement to associate each node ID with an output
   *output_switch_statements = "switch (nid) {\n";
-  for (OutputNode* e : output_nodes) {
+  for (OutputNode<LeafOutputType>* e : output_nodes) {
     const int node_id = descendants[static_cast<ASTNode*>(e)];
     *output_switch_statements
       += fmt::format(" case {node_id}:\n"
